@@ -10,10 +10,9 @@
 */
 
 #include "kcmultidialog.h"
-#include "kcmultidialog_p.h"
-
-#include "kcmoduleproxy.h"
+#include "kcmoduleloader.h"
 #include "kcmoduleqml_p.h"
+#include "kcmultidialog_p.h"
 #include <kcmutils_debug.h>
 
 #include <QApplication>
@@ -26,22 +25,18 @@
 #include <QStandardPaths>
 #include <QStringList>
 #include <QStyle>
+#include <QTimer>
 #include <QUrl>
 
-#if KCONFIGWIDGETS_WITH_KAUTH
-#include <KAuth/Action>
-#include <KAuth/ObjectDecorator>
-#endif
 #include <KGuiItem>
 #include <KIconUtils>
 #include <KLocalizedString>
 #include <KMessageBox>
 #include <KPageWidgetModel>
 
-bool KCMultiDialogPrivate::resolveChanges(KCModuleProxy *currentProxy)
+bool KCMultiDialogPrivate::resolveChanges(KCModule *module)
 {
-    Q_Q(KCMultiDialog);
-    if (!currentProxy || !currentProxy->isChanged()) {
+    if (!module || !module->needsSave()) {
         return true;
     }
 
@@ -56,10 +51,10 @@ bool KCMultiDialogPrivate::resolveChanges(KCModuleProxy *currentProxy)
 
     switch (queryUser) {
     case KMessageBox::PrimaryAction:
-        return moduleSave(currentProxy);
+        return moduleSave(module);
 
     case KMessageBox::SecondaryAction:
-        currentProxy->load();
+        module->load();
         return true;
 
     case KMessageBox::Cancel:
@@ -71,11 +66,9 @@ bool KCMultiDialogPrivate::resolveChanges(KCModuleProxy *currentProxy)
     }
 }
 
-void KCMultiDialogPrivate::_k_slotCurrentPageChanged(KPageWidgetItem *current, KPageWidgetItem *previous)
+void KCMultiDialogPrivate::slotCurrentPageChanged(KPageWidgetItem *current, KPageWidgetItem *previous)
 {
-    Q_Q(KCMultiDialog);
-
-    KCModuleProxy *previousModule = nullptr;
+    KCModule *previousModule = nullptr;
     for (int i = 0; i < modules.count(); ++i) {
         if (modules[i].item == previous) {
             previousModule = modules[i].kcm;
@@ -112,74 +105,39 @@ void KCMultiDialogPrivate::_k_slotCurrentPageChanged(KPageWidgetItem *current, K
     q->blockSignals(false);
 
     // We need to get the state of the now active module
-    _k_clientChanged();
+    clientChanged();
 }
 
-void KCMultiDialogPrivate::_k_clientChanged()
+void KCMultiDialogPrivate::clientChanged()
 {
-    Q_Q(KCMultiDialog);
-    // qDebug();
     // Get the current module
-    KCModuleProxy *activeModule = nullptr;
+    KCModule *activeModule = nullptr;
+    bool scheduleFirstShow = false;
     for (int i = 0; i < modules.count(); ++i) {
         if (modules[i].item == q->currentPage()) {
             activeModule = modules[i].kcm;
+            scheduleFirstShow = activeModule && modules[i].firstShow;
             break;
         }
     }
 
-    bool change = false;
-    bool defaulted = false;
-    if (activeModule) {
-        change = activeModule->isChanged();
-        defaulted = activeModule->defaulted();
-
-        QPushButton *applyButton = q->buttonBox()->button(QDialogButtonBox::Apply);
-        if (applyButton) {
-            q->disconnect(applyButton, &QAbstractButton::clicked, q, &KCMultiDialog::slotApplyClicked);
-#if KCONFIGWIDGETS_WITH_KAUTH
-            delete applyButton->findChild<KAuth::ObjectDecorator *>();
-#endif
-        }
-
-        QPushButton *okButton = q->buttonBox()->button(QDialogButtonBox::Ok);
-        if (okButton) {
-            q->disconnect(okButton, &QAbstractButton::clicked, q, &KCMultiDialog::slotOkClicked);
-#if KCONFIGWIDGETS_WITH_KAUTH
-            delete okButton->findChild<KAuth::ObjectDecorator *>();
-#endif
-        }
-
-#if KCONFIGWIDGETS_WITH_KAUTH
-        if (activeModule->realModule()->needsAuthorization()) {
-            if (applyButton) {
-                KAuth::ObjectDecorator *decorator = new KAuth::ObjectDecorator(applyButton);
-                decorator->setAuthAction(activeModule->realModule()->authAction());
-                activeModule->realModule()->authAction().setParentWidget(activeModule->realModule());
-                q->connect(decorator, &KAuth::ObjectDecorator::authorized, q, &KCMultiDialog::slotApplyClicked);
+    // When we first show a module, we call the load method
+    // Just in case we have multiple loadModule calls in a row, the current module could change
+    // Meaning we wait for the next tick, check the active module and call load if needed
+    if (scheduleFirstShow) {
+        QTimer::singleShot(0, q, [this]() {
+            for (int i = 0; i < modules.count(); ++i) {
+                if (modules[i].firstShow && modules[i].kcm && modules[i].item == q->currentPage()) {
+                    modules[i].kcm->load();
+                    modules[i].firstShow = false;
+                }
             }
-
-            if (okButton) {
-                KAuth::ObjectDecorator *decorator = new KAuth::ObjectDecorator(okButton);
-                decorator->setAuthAction(activeModule->realModule()->authAction());
-                activeModule->realModule()->authAction().setParentWidget(activeModule->realModule());
-                q->connect(decorator, &KAuth::ObjectDecorator::authorized, q, &KCMultiDialog::slotOkClicked);
-            }
-        } else {
-            if (applyButton) {
-                q->connect(applyButton, &QAbstractButton::clicked, q, &KCMultiDialog::slotApplyClicked);
-                delete applyButton->findChild<KAuth::ObjectDecorator *>();
-            }
-
-            if (okButton) {
-                q->connect(okButton, &QAbstractButton::clicked, q, &KCMultiDialog::slotOkClicked);
-                delete okButton->findChild<KAuth::ObjectDecorator *>();
-            }
-        }
-#endif
+        });
     }
 
-    auto buttons = activeModule ? activeModule->buttons() : KCModule::NoAdditionalButton;
+    const bool change = activeModule && activeModule->needsSave();
+    const bool defaulted = activeModule && activeModule->representsDefaults();
+    const auto buttons = activeModule ? activeModule->buttons() : KCModule::NoAdditionalButton;
 
     QPushButton *resetButton = q->buttonBox()->button(QDialogButtonBox::Reset);
     if (resetButton) {
@@ -220,26 +178,17 @@ void KCMultiDialogPrivate::_k_clientChanged()
     }
 }
 
-void KCMultiDialogPrivate::_k_updateHeader(bool use, const QString &message)
+void KCMultiDialogPrivate::updateHeader(bool use, const QString &message)
 {
-    Q_Q(KCMultiDialog);
     KPageWidgetItem *item = q->currentPage();
-    KCModuleProxy *kcm = qobject_cast<KCModuleProxy *>(item->widget());
+    const auto findIt = std::find_if(modules.cbegin(), modules.cend(), [item](const CreatedModule &module) {
+        return module.item == item;
+    });
+    Q_ASSERT(findIt != modules.cend());
 
-    QString moduleName;
-    QString icon;
-
-    if (kcm->metaData().isValid()) {
-        moduleName = kcm->metaData().name();
-        icon = kcm->metaData().iconName();
-    }
-
-#if KCMUTILS_BUILD_DEPRECATED_SINCE(5, 88)
-    if (kcm->moduleInfo().isValid()) {
-        moduleName = kcm->moduleInfo().moduleName();
-        icon = kcm->moduleInfo().icon();
-    }
-#endif
+    KCModule *kcm = findIt->kcm;
+    const QString moduleName = kcm->metaData().name();
+    const QString icon = kcm->metaData().iconName();
 
     if (use) {
         item->setHeader(QStringLiteral("<b>") + moduleName + QStringLiteral("</b><br><i>") + message + QStringLiteral("</i>"));
@@ -252,7 +201,6 @@ void KCMultiDialogPrivate::_k_updateHeader(bool use, const QString &message)
 
 void KCMultiDialogPrivate::init()
 {
-    Q_Q(KCMultiDialog);
     q->setFaceType(KPageDialog::Auto);
     q->setWindowTitle(i18n("Configure"));
     q->setModal(false);
@@ -279,35 +227,18 @@ void KCMultiDialogPrivate::init()
 
     q->setButtonBox(buttonBox);
     q->connect(q, &KPageDialog::currentPageChanged, q, [this](KPageWidgetItem *current, KPageWidgetItem *before) {
-        _k_slotCurrentPageChanged(current, before);
+        slotCurrentPageChanged(current, before);
     });
 }
 
 KCMultiDialog::KCMultiDialog(QWidget *parent)
     : KPageDialog(parent)
-    , d_ptr(new KCMultiDialogPrivate(this))
+    , d(new KCMultiDialogPrivate(this))
 {
-    d_func()->init();
+    d->init();
 }
 
-KCMultiDialog::KCMultiDialog(KPageWidget *pageWidget, QWidget *parent, Qt::WindowFlags flags)
-    : KPageDialog(pageWidget, parent, flags)
-    , d_ptr(new KCMultiDialogPrivate(this))
-{
-    d_func()->init();
-}
-
-KCMultiDialog::KCMultiDialog(KCMultiDialogPrivate &dd, KPageWidget *pageWidget, QWidget *parent, Qt::WindowFlags flags)
-    : KPageDialog(pageWidget, parent, flags)
-    , d_ptr(&dd)
-{
-    d_func()->init();
-}
-
-KCMultiDialog::~KCMultiDialog()
-{
-    delete d_ptr;
-}
+KCMultiDialog::~KCMultiDialog() = default;
 
 void KCMultiDialog::showEvent(QShowEvent *ev)
 {
@@ -327,7 +258,6 @@ void KCMultiDialog::showEvent(QShowEvent *ev)
 
 void KCMultiDialog::slotDefaultClicked()
 {
-    Q_D(KCMultiDialog);
     const KPageWidgetItem *item = currentPage();
     if (!item) {
         return;
@@ -336,7 +266,7 @@ void KCMultiDialog::slotDefaultClicked()
     for (int i = 0; i < d->modules.count(); ++i) {
         if (d->modules[i].item == item) {
             d->modules[i].kcm->defaults();
-            d->_k_clientChanged();
+            d->clientChanged();
             return;
         }
     }
@@ -349,17 +279,16 @@ void KCMultiDialog::slotUser1Clicked()
         return;
     }
 
-    Q_D(KCMultiDialog);
     for (int i = 0; i < d->modules.count(); ++i) {
         if (d->modules[i].item == item) {
             d->modules[i].kcm->load();
-            d->_k_clientChanged();
+            d->clientChanged();
             return;
         }
     }
 }
 
-bool KCMultiDialogPrivate::moduleSave(KCModuleProxy *module)
+bool KCMultiDialogPrivate::moduleSave(KCModule *module)
 {
     if (!module) {
         return false;
@@ -371,36 +300,14 @@ bool KCMultiDialogPrivate::moduleSave(KCModuleProxy *module)
 
 void KCMultiDialogPrivate::apply()
 {
-    Q_Q(KCMultiDialog);
-    QStringList updatedComponents;
-
     for (const CreatedModule &module : std::as_const(modules)) {
-        KCModuleProxy *proxy = module.kcm;
+        KCModule *kcm = module.kcm;
 
-        if (proxy->isChanged()) {
-            proxy->save();
-#if KCMUTILS_BUILD_DEPRECATED_SINCE(5, 85)
-            /**
-             * Add name of the components the kcm belongs to the list
-             * of updated components.
-             */
-            const QStringList componentNames = module.componentNames;
-            for (const QString &componentName : componentNames) {
-                if (!updatedComponents.contains(componentName)) {
-                    updatedComponents.append(componentName);
-                }
-            }
-#endif
+        if (kcm->needsSave()) {
+            kcm->save();
         }
     }
 
-#if KCMUTILS_BUILD_DEPRECATED_SINCE(5, 85)
-    // Send the configCommitted signal for every updated component.
-    for (const QString &name : std::as_const(updatedComponents)) {
-        Q_EMIT q->configCommitted(name.toLatin1());
-    }
-#endif
-
     Q_EMIT q->configCommitted();
 }
 
@@ -409,7 +316,7 @@ void KCMultiDialog::slotApplyClicked()
     QPushButton *applyButton = buttonBox()->button(QDialogButtonBox::Apply);
     applyButton->setFocus();
 
-    d_func()->apply();
+    d->apply();
 }
 
 void KCMultiDialog::slotOkClicked()
@@ -417,7 +324,7 @@ void KCMultiDialog::slotOkClicked()
     QPushButton *okButton = buttonBox()->button(QDialogButtonBox::Ok);
     okButton->setFocus();
 
-    d_func()->apply();
+    d->apply();
     accept();
 }
 
@@ -428,13 +335,9 @@ void KCMultiDialog::slotHelpClicked()
         return;
     }
 
-    Q_D(KCMultiDialog);
     QString docPath;
     for (int i = 0; i < d->modules.count(); ++i) {
         if (d->modules[i].item == item) {
-#if KCMUTILS_BUILD_DEPRECATED_SINCE(5, 88)
-            docPath = d->modules[i].kcm->moduleInfo().docPath();
-#endif
             if (docPath.isEmpty()) {
                 docPath = d->modules[i].kcm->metaData().value(QStringLiteral("X-DocPath"));
             }
@@ -458,27 +361,16 @@ void KCMultiDialog::slotHelpClicked()
 
 void KCMultiDialog::closeEvent(QCloseEvent *event)
 {
-    Q_D(KCMultiDialog);
     KPageDialog::closeEvent(event);
 
-    /**
-     * If we don't delete them, the DBUS registration stays, and trying to load the KCMs
-     * in other situations will lead to "module already loaded in Foo," while to the user
-     * doesn't appear so(the dialog is hidden)
-     */
-    for (auto &proxy : std::as_const(d->modules)) {
-        proxy.kcm->deleteClient();
+    for (auto &module : d->modules) {
+        delete module.kcm;
+        module.kcm = nullptr;
     }
 }
 
-KPageWidgetItem *KCMultiDialog::addModule(const KPluginMetaData &metaData)
+KPageWidgetItem *KCMultiDialog::addModule(const KPluginMetaData &metaData, const QVariantList &args)
 {
-    return addModule(metaData, QStringList());
-}
-
-KPageWidgetItem *KCMultiDialog::addModule(const KPluginMetaData &metaData, const QStringList &args)
-{
-    Q_D(KCMultiDialog);
     // Create the scroller
     auto *moduleScroll = new UnboundScrollArea(this);
     // Prepare the scroll area
@@ -486,8 +378,8 @@ KPageWidgetItem *KCMultiDialog::addModule(const KPluginMetaData &metaData, const
     moduleScroll->setFrameStyle(QFrame::NoFrame);
     moduleScroll->viewport()->setAutoFillBackground(false);
 
-    KCModuleProxy *kcm = new KCModuleProxy(metaData, moduleScroll, args);
-    moduleScroll->setWidget(kcm);
+    KCModule *kcm = KCModuleLoader::loadModule(metaData, moduleScroll, args);
+    moduleScroll->setWidget(kcm->widget());
 
     KPageWidgetItem *item = new KPageWidgetItem(moduleScroll, metaData.name());
 
@@ -496,17 +388,12 @@ KPageWidgetItem *KCMultiDialog::addModule(const KPluginMetaData &metaData, const
     createdModule.item = item;
     d->modules.append(createdModule);
 
-    if (qobject_cast<KCModuleQml *>(kcm->realModule())) {
+    if (qobject_cast<KCModuleQml *>(kcm)) {
         item->setHeaderVisible(false);
     }
 
-    if (kcm->realModule() && kcm->realModule()->useRootOnlyMessage()) {
-        item->setHeader(QStringLiteral("<b>%1</b><br><i>%2</i>").arg(metaData.name(), kcm->realModule()->rootOnlyMessage()));
-        item->setIcon(KIconUtils::addOverlay(QIcon::fromTheme(metaData.iconName()), QIcon::fromTheme(QStringLiteral("dialog-warning")), Qt::BottomRightCorner));
-    } else {
-        item->setHeader(metaData.name());
-        item->setIcon(QIcon::fromTheme(metaData.iconName()));
-    }
+    item->setHeader(metaData.name());
+    item->setIcon(QIcon::fromTheme(metaData.iconName()));
     const int weight = metaData.rawData().value(QStringLiteral("X-KDE-Weight")).toInt();
     item->setProperty("_k_weight", weight);
 
@@ -534,155 +421,33 @@ KPageWidgetItem *KCMultiDialog::addModule(const KPluginMetaData &metaData, const
         addPage(item);
     }
 
-    QObject::connect(kcm, qOverload<bool>(&KCModuleProxy::changed), this, [d]() {
-        d->_k_clientChanged();
-    });
-
-    QObject::connect(kcm->realModule(), &KCModule::rootOnlyMessageChanged, this, [d](bool use, const QString &message) {
-        d->_k_updateHeader(use, message);
+    connect(kcm, &KCModule::needsSaveChanged, this, [this]() {
+        d->clientChanged();
     });
 
     if (d->modules.count() == 1 || updateCurrentPage) {
         setCurrentPage(item);
-        d->_k_clientChanged();
+        d->clientChanged();
     }
     return item;
 }
 
-#if KCMUTILS_BUILD_DEPRECATED_SINCE(5, 85)
-KPageWidgetItem *KCMultiDialog::addModule(const QString &path, const QStringList &args)
-{
-    QString complete = path;
-
-    if (!path.endsWith(QLatin1String(".desktop"))) {
-        complete += QStringLiteral(".desktop");
-    }
-
-    KService::Ptr service = KService::serviceByStorageId(complete);
-
-    return addModule(KCModuleInfo(service), nullptr, args);
-}
-#endif
-
-#if KCMUTILS_BUILD_DEPRECATED_SINCE(5, 85)
-KPageWidgetItem *KCMultiDialog::addModule(const KCModuleInfo &moduleInfo, KPageWidgetItem *parentItem, const QStringList &args)
-{
-    Q_D(KCMultiDialog);
-    if (!moduleInfo.isValid()) {
-        return nullptr;
-    }
-
-    // KAuthorized::authorizeControlModule( moduleInfo.service()->menuId() ) is
-    // checked in noDisplay already
-    if (moduleInfo.service() && moduleInfo.service()->noDisplay()) {
-        return nullptr;
-    }
-
-    // Create the scroller
-    auto *moduleScroll = new UnboundScrollArea(this);
-    // Prepare the scroll area
-    moduleScroll->setWidgetResizable(true);
-    moduleScroll->setFrameStyle(QFrame::NoFrame);
-    moduleScroll->viewport()->setAutoFillBackground(false);
-
-    KCModuleProxy *kcm = new KCModuleProxy(moduleInfo, moduleScroll, args);
-    moduleScroll->setWidget(kcm);
-
-    // qDebug() << moduleInfo.moduleName();
-    KPageWidgetItem *item = new KPageWidgetItem(moduleScroll, moduleInfo.moduleName());
-
-    KCMultiDialogPrivate::CreatedModule cm;
-    cm.kcm = kcm;
-    cm.item = item;
-    cm.componentNames = moduleInfo.property(QStringLiteral("X-KDE-ParentComponents")).toStringList();
-    d->modules.append(cm);
-
-    if (qobject_cast<KCModuleQml *>(kcm->realModule())) {
-        item->setHeaderVisible(false);
-    }
-
-    if (kcm->realModule() && kcm->realModule()->useRootOnlyMessage()) {
-        item->setHeader(QStringLiteral("<b>") + moduleInfo.moduleName() + QStringLiteral("</b><br><i>") + kcm->realModule()->rootOnlyMessage()
-                        + QStringLiteral("</i>"));
-        item->setIcon(KIconUtils::addOverlay(QIcon::fromTheme(moduleInfo.icon()), QIcon::fromTheme(QStringLiteral("dialog-warning")), Qt::BottomRightCorner));
-    } else {
-        item->setHeader(moduleInfo.moduleName());
-        item->setIcon(QIcon::fromTheme(moduleInfo.icon()));
-    }
-    item->setProperty("_k_weight", moduleInfo.weight());
-
-    bool updateCurrentPage = false;
-    const KPageWidgetModel *model = qobject_cast<const KPageWidgetModel *>(pageWidget()->model());
-    Q_ASSERT(model);
-    if (parentItem) {
-        const QModelIndex parentIndex = model->index(parentItem);
-        const int siblingCount = model->rowCount(parentIndex);
-        int row = 0;
-        for (; row < siblingCount; ++row) {
-            KPageWidgetItem *siblingItem = model->item(model->index(row, 0, parentIndex));
-            if (siblingItem->property("_k_weight").toInt() > moduleInfo.weight()) {
-                // the item we found is heavier than the new module
-                // qDebug() << "adding KCM " << item->name() << " before " << siblingItem->name();
-                insertPage(siblingItem, item);
-                break;
-            }
-        }
-        if (row >= siblingCount) {
-            // the new module is either the first or the heaviest item
-            // qDebug() << "adding KCM " << item->name() << " with parent " << parentItem->name();
-            addSubPage(parentItem, item);
-        }
-    } else {
-        const int siblingCount = model->rowCount();
-        int row = 0;
-        for (; row < siblingCount; ++row) {
-            KPageWidgetItem *siblingItem = model->item(model->index(row, 0));
-            if (siblingItem->property("_k_weight").toInt() > moduleInfo.weight()) {
-                // the item we found is heavier than the new module
-                // qDebug() << "adding KCM " << item->name() << " before " << siblingItem->name();
-                insertPage(siblingItem, item);
-                if (siblingItem == currentPage()) {
-                    updateCurrentPage = true;
-                }
-
-                break;
-            }
-        }
-        if (row == siblingCount) {
-            // the new module is either the first or the heaviest item
-            // qDebug() << "adding KCM " << item->name() << " at the top level";
-            addPage(item);
-        }
-    }
-
-    QObject::connect(kcm, qOverload<bool>(&KCModuleProxy::changed), this, [d]() {
-        d->_k_clientChanged();
-    });
-
-    QObject::connect(kcm->realModule(), &KCModule::rootOnlyMessageChanged, this, [d](bool use, QString message) {
-        d->_k_updateHeader(use, message);
-    });
-
-    if (d->modules.count() == 1 || updateCurrentPage) {
-        setCurrentPage(item);
-        d->_k_clientChanged();
-    }
-    return item;
-}
-#endif
-
 void KCMultiDialog::clear()
 {
-    Q_D(KCMultiDialog);
-    // qDebug() ;
-
     for (int i = 0; i < d->modules.count(); ++i) {
         removePage(d->modules[i].item);
     }
 
     d->modules.clear();
 
-    d->_k_clientChanged();
+    d->clientChanged();
+}
+
+void KCMultiDialog::setDefaultsIndicatorsVisible(bool show)
+{
+    for (const auto &module : std::as_const(d->modules)) {
+        module.kcm->setDefaultsIndicatorsVisible(show);
+    }
 }
 
 #include "moc_kcmultidialog.cpp"
