@@ -6,6 +6,7 @@
 */
 
 #include "kpluginwidget.h"
+#include "kcmoduleloader.h"
 #include "kpluginproxymodel.h"
 #include "kpluginwidget_p.h"
 
@@ -32,8 +33,6 @@
 #include <KLocalizedString>
 #include <KPluginMetaData>
 #include <KStandardGuiItem>
-#include <kcmoduleinfo.h>
-#include <kcmoduleproxy.h>
 #include <utility>
 
 static constexpr int s_margin = 5;
@@ -69,7 +68,7 @@ KPluginWidget::KPluginWidget(QWidget *parent)
     connect(d->pluginModel,
             &QAbstractItemModel::dataChanged,
             this,
-            [this](const QModelIndex &topLeft, const QModelIndex & /*bottomRight*/, const QVector<int> &roles) {
+            [this](const QModelIndex &topLeft, const QModelIndex & /*bottomRight*/, const QList<int> &roles) {
                 if (roles.contains(KPluginModel::EnabledRole)) {
                     Q_EMIT pluginEnabledChanged(topLeft.data(KPluginModel::IdRole).toString(), topLeft.data(KPluginModel::EnabledRole).toBool());
                     Q_EMIT changed(d->pluginModel->isSaveNeeded());
@@ -77,7 +76,7 @@ KPluginWidget::KPluginWidget(QWidget *parent)
             });
 
     d->proxyModel = new KPluginProxyModel(this);
-    d->proxyModel->setSourceModel(d->pluginModel);
+    d->proxyModel->setModel(d->pluginModel);
     d->listView->setModel(d->proxyModel);
     d->listView->setAlternatingRowColors(true);
 
@@ -108,7 +107,7 @@ KPluginWidget::~KPluginWidget()
     delete d->listView; // depends on some other things in d, make sure this dies first.
 }
 
-void KPluginWidget::addPlugins(const QVector<KPluginMetaData> &plugins, const QString &categoryLabel)
+void KPluginWidget::addPlugins(const QList<KPluginMetaData> &plugins, const QString &categoryLabel)
 {
     d->pluginModel->addPlugins(plugins, categoryLabel);
     d->proxyModel->sort(0);
@@ -156,12 +155,12 @@ bool KPluginWidget::isSaveNeeded() const
     return d->pluginModel->isSaveNeeded();
 }
 
-void KPluginWidget::setConfigurationArguments(const QStringList &arguments)
+void KPluginWidget::setConfigurationArguments(const QVariantList &arguments)
 {
     d->kcmArguments = arguments;
 }
 
-QStringList KPluginWidget::configurationArguments() const
+QVariantList KPluginWidget::configurationArguments() const
 {
     return d->kcmArguments;
 }
@@ -354,7 +353,7 @@ QList<QWidget *> PluginDelegate::createItemWidgets(const QModelIndex &index) con
     return widgetList;
 }
 
-void PluginDelegate::updateItemWidgets(const QList<QWidget *> widgets, const QStyleOptionViewItem &option, const QPersistentModelIndex &index) const
+void PluginDelegate::updateItemWidgets(const QList<QWidget *> &widgets, const QStyleOptionViewItem &option, const QPersistentModelIndex &index) const
 {
     int extraButtonWidth = 0;
     QPushButton *extraButton = nullptr;
@@ -446,15 +445,11 @@ void PluginDelegate::configure(const QModelIndex &index)
     configDialog->setModal(true);
     configDialog->setWindowTitle(model->data(index, KPluginModel::NameRole).toString());
 
-    auto moduleProxy = new KCModuleProxy(kcm, configDialog, pluginSelector_d->kcmArguments);
-
-    if (!moduleProxy->realModule()) {
-        delete moduleProxy;
-        return;
-    }
+    QWidget *kcmWrapper = new QWidget;
+    auto kcmInstance = KCModuleLoader::loadModule(kcm, kcmWrapper, pluginSelector_d->kcmArguments);
 
     auto layout = new QVBoxLayout(configDialog);
-    layout->addWidget(moduleProxy);
+    layout->addWidget(kcmWrapper);
 
     auto buttonBox = new QDialogButtonBox(configDialog);
     buttonBox->setStandardButtons(QDialogButtonBox::Ok | QDialogButtonBox::Cancel | QDialogButtonBox::RestoreDefaults);
@@ -463,19 +458,21 @@ void PluginDelegate::configure(const QModelIndex &index)
     KGuiItem::assign(buttonBox->button(QDialogButtonBox::RestoreDefaults), KStandardGuiItem::defaults());
     connect(buttonBox, &QDialogButtonBox::accepted, configDialog, &QDialog::accept);
     connect(buttonBox, &QDialogButtonBox::rejected, configDialog, &QDialog::reject);
-    connect(configDialog, &QDialog::accepted, this, [moduleProxy, this, model, index]() {
+    connect(configDialog, &QDialog::accepted, this, [kcmInstance, this, model, index]() {
         Q_EMIT configCommitted(model->data(index, KPluginModel::IdRole).toString());
-        moduleProxy->save();
+        kcmInstance->save();
     });
-    connect(configDialog, &QDialog::rejected, this, [moduleProxy]() {
-        moduleProxy->load();
+    connect(configDialog, &QDialog::rejected, this, [kcmInstance]() {
+        kcmInstance->load();
     });
 
-    connect(buttonBox->button(QDialogButtonBox::RestoreDefaults), &QAbstractButton::clicked, this, [moduleProxy] {
-        moduleProxy->defaults();
+    connect(buttonBox->button(QDialogButtonBox::RestoreDefaults), &QAbstractButton::clicked, this, [kcmInstance] {
+        kcmInstance->defaults();
     });
     layout->addWidget(buttonBox);
 
+    // Load KCM right before showing it
+    kcmInstance->load();
     configDialog->show();
 }
 
